# -*- coding: utf-8 -*-

from twisted.application import internet
from twisted.application.service import IService, Service, MultiService
from twisted.python import log
from zope.interface import implementer, Interface

from il2ds_middleware.ds_emulator.interfaces import ILineBroadcaster


class ILineParser(Interface):

    def parse_line(self, line):
        """
        Return True if line was successfully parsed, otherwise return False.
        """
        raise NotImplementedError


class _PropagatorMixin:

    propagate = False

    def _autopropagate(self, value=True):
        return False if self.propagate else value


@implementer(IService, ILineBroadcaster)
class _LineBroadcastingServiceMixin:

    broadcaster = None

    def broadcast_line(self, line):
        if self.broadcaster:
            self.broadcaster.broadcast_line(line)
        elif self.parent:
            self.parent.broadcast_line(line)
        else:
            log.msg("Broadcasting into nowhere: \"{0}\"".format(line))


@implementer(ILineParser)
class _DSServiceMixin(_LineBroadcastingServiceMixin, _PropagatorMixin):
    pass


class RootService(MultiService, _DSServiceMixin):

    """
    Top-level service.
    """

    def __init__(self, broadcaster):
        MultiService.__init__(self)
        self.broadcaster = broadcaster
        self._init_children()

    def _init_children(self):
        """
        Initialize children services.
        """
        PilotService().setServiceParent(self)
        MissionService().setServiceParent(self)

    def startService(self):
        self.broadcaster.service = self
        MultiService.startService(self)

    def stopService(self):
        self.broadcaster.service = None
        return MultiService.stopService(self)

    def parse_line(self, line):
        result = False
        for service in self.services:
            if not ILineParser.providedBy(service):
                continue
            result = service.parse_line(line)
            if result:
                break
        if not result:
            self.broadcast_line("Command not found: " + line)
        return self._autopropagate(result)


class PilotService(Service, _DSServiceMixin):

    name = "pilots"
    channel = 1
    channel_inc = 2
    port = 21000

    def __init__(self):
        self.pilots = []

    def parse_line(self, line):
        while True:
            if line.startswith("kick"):
                self._kick(callsign=line[4:].strip())
                break
            return self._autopropagate(False)
        return self._autopropagate()

    def join(self, callsign, ip):

        def create_pilot():
            pilot = {
                'callsign': callsign,
                'ip': ip,
                'channel': self.channel,
            }
            self.channel += self.channel_inc
            return pilot

        pilot = create_pilot()
        self.pilots.append(pilot)

        self.broadcast_line(
            "socket channel '{0}' start creating: ip {1}:{2}".format(
                pilot['channel'], pilot['ip'], self.port))
        self.broadcast_line(
            "Chat: --- {0} joins the game.".format(
                pilot['callsign']))
        self.broadcast_line(
            "socket channel '{0}', ip {1}:{2}, {3}, "
            "is complete created.".format(
                pilot['channel'], pilot['ip'], self.port, pilot['callsign']))

    def leave(self, callsign):
        self._leave(callsign)

    def _kick(self, callsign):
        self._leave(callsign, reason="You have been kicked from the server.")

    def _leave(self, callsign, reason=None):
        pilot = self._pilot_by_callsign(callsign)
        if pilot is None:
            return
        self.pilots.remove(pilot)

        line = "socketConnection with {0}:{1} on channel {2} lost.  " \
            "Reason: ".format(pilot['ip'], self.port, pilot['channel'])
        if reason:
            line += reason
        self.broadcast_line(line)
        self.broadcast_line("Chat: --- {0} has left the game.".format(
            pilot['callsign']))

    def _pilot_by_callsign(self, callsign):
        for p in self.pilots:
            if p['callsign'] == callsign:
                return p
        log.err("Pilot with callsign \"{0}\" not found.".format(callsign))
        return None

    def stopService(self):
        self.pilots = None
        return Service.stopService(self)


MISSION_NONE, MISSION_LOADED, MISSION_PLAYING = tuple(range(1, 4))


class MissionService(Service, _DSServiceMixin):

    name = "missions"
    status = MISSION_NONE
    mission = None

    def parse_line(self, line):
        if not line.startswith("mission"):
            return self._autopropagate(False)
        cmd = line[7:].strip()
        while True:
            if not cmd:
                self._send_status()
                break
            elif cmd.startswith("LOAD"):
                self._load_mission(mission=cmd[4:].lstrip())
                break
            elif cmd == "BEGIN":
                self._begin_mission()
                break
            elif cmd == "END":
                self._end_mission()
                break
            elif cmd == "DESTROY":
                self._destroy_mission()
                break
            return self._autopropagate(False)
        return self._autopropagate()

    def _load_mission(self, mission):
        self.mission = mission
        self.broadcast_line("Loading mission {0}...".format(self.mission))
        self.broadcast_line("Load bridges")
        self.broadcast_line("Load static objects")
        self.broadcast_line("##### House without collision "
            "(3do/Tree/Tree2.sim)")
        self.broadcast_line("##### House without collision "
            "(3do/Buildings/Port/Floor/live.sim)")
        self.broadcast_line("##### House without collision "
            "(3do/Buildings/Port/BaseSegment/live.sim)")
        self.status = MISSION_LOADED
        self._send_status()

    def _begin_mission(self):
        if self.status == MISSION_NONE:
            self._mission_not_loaded()
        else:
            self.status = MISSION_PLAYING
            self._send_status()

    def _end_mission(self):
        if self.status == MISSION_NONE:
            self._mission_not_loaded()
        else:
            self.status = MISSION_LOADED
            self._send_status()

    def _destroy_mission(self):
        if self.status == MISSION_NONE:
            self._mission_not_loaded()
        else:
            self.status = MISSION_NONE
            self.mission = None

    def _mission_not_loaded(self):
        self.broadcast_line("ERROR mission: Mission NOT loaded")

    def _send_status(self):
        if self.status == MISSION_NONE:
            self.broadcast_line("Mission NOT loaded")
        elif self.status == MISSION_LOADED:
            self.broadcast_line("Mission: {0} is Loaded".format(self.mission))
        elif self.status == MISSION_PLAYING:
            self.broadcast_line("Mission: {0} is Playing".format(self.mission))

    def stopService(self):
        self.mission = None
        return Service.stopService(self)



# -*- coding: utf-8 -*-

from twisted.internet import reactor
from twisted.internet.protocol import ClientFactory
from twisted.protocols.basic import LineOnlyReceiver
from twisted.python import log


class ConsoleProtocol(LineOnlyReceiver):

    def connectionMade(self):
        log.msg("Connection established with {0}".format(
            self.transport.getPeer()))

    def connectionLost(self, reason):
        log.err("Connection with {0} lost: {1}".format(
            self.transport.getPeer(), reason))

    def lineReceived(self, line):
        if line == '' or line == "exit\\n":
            log.err("Game server is shut down")
            reactor.stop()
            return
        if line.startswith("<consoleN>"):
            return
        if line.endswith("\\n"):
            line = line[:-2]
        if line.startswith("\\u0020"):
            line = " " + line[6:]
        # TODO:


class ConsoleFactory(ClientFactory):

    protocol = ConsoleProtocol

    def clientConnectionFailed(self, connector, reason):
        log.err("Connection failed: %s" % reason)



# -*- coding: utf-8 -*-

from twisted.internet import defer, protocol
from twisted.protocols.basic import LineReceiver
from twisted.trial.unittest import FailTest, TestCase

from il2ds_middleware.ds_emulator.service import RootService
from il2ds_middleware.ds_emulator.protocol import (ConsoleFactory,
    ConsoleProtocol)


class ServerFactory(ConsoleFactory):

    def __init__(self):
        ConsoleFactory.__init__(self)
        self.on_connection_lost = defer.Deferred()

    def client_left(self, client):
        ConsoleFactory.client_left(self, client)
        self.on_connection_lost.callback(client)


class ClientProtocol(LineReceiver):

    def connectionMade(self):
        self.factory.client_joined(self)

    def connectionLost(self, reason):
        self.factory.client_left(self)

    def lineReceived(self, line):
        if self.factory.receiver:
            self.factory.receiver(line)


class ClientFactory(protocol.ClientFactory):

    protocol = ClientProtocol

    def __init__(self):
        self.clients = []
        self.on_connection_made = defer.Deferred()
        self.on_connection_lost = defer.Deferred()

    def client_joined(self, client):
        self.on_connection_made.callback(client)
        self.clients.append(client)

    def client_left(self, client):
        self.on_connection_lost.callback(client)
        self.clients.remove(client)

    def message(self, message):

        def do_message(message):
            for client in self.clients:
                client.sendLine(message)

        from twisted.internet import reactor
        reactor.callLater(0, do_message, message)


class ConsoleBaseTestCase(TestCase):

    def setUp(self):
        self.server_port = self._listen_server()
        self.client_connection, connected = self._connect_client()
        return connected

    def _listen_server(self):
        self.sfactory = ServerFactory()
        self.sservice = RootService(self.sfactory)
        self.sfactory.service = self.sservice
        self.sservice.startService()

        from twisted.internet import reactor
        return reactor.listenTCP(0, self.sfactory, interface="127.0.0.1")

    def _connect_client(self):
        self.cfactory = ClientFactory()
        from twisted.internet import reactor
        return (reactor.connectTCP(
            "127.0.0.1", self.server_port.getHost().port, self.cfactory),
            self.cfactory.on_connection_made)

    def tearDown(self):
        listening_stopped = defer.maybeDeferred(self.server_port.stopListening)
        self.client_connection.disconnect()
        self.cfactory.receiver = None
        return defer.gatherResults([
            listening_stopped,
            self.cfactory.on_connection_lost, self.sfactory.on_connection_lost,
            self.sservice.stopService()])

    def _get_expecting_line_receiver(self, expected_responses, d):

        def got_line(line):
            try:
                self.assertEqual(line, expected_responses.pop(0))
            except Exception as e:
                timeout.cancel()
                d.errback(e)
            else:
                if expected_responses:
                    return
                timeout.cancel()
                d.callback(None)

        def on_timeout(_):
            d.errback(FailTest(
            'Timed out, remaining lines:\n\t'+'\n\t'.join(expected_responses)))

        from twisted.internet import reactor
        timeout = reactor.callLater(0.1, on_timeout, None)
        return got_line


class TestConnection(ConsoleBaseTestCase):

    def test_connect(self):
        self.assertEqual(len(self.sfactory.clients), 1)

    def test_disconnect(self):
        self.sfactory.on_connection_lost.addBoth(
            lambda _: self.assertEqual(len(self.sfactory.clients), 0))
        self.client_connection.disconnect()

    def test_receive_line(self):
        d = defer.Deferred()
        responses = ["test\\n", ]
        self.cfactory.receiver = self._get_expecting_line_receiver(
            responses, d)
        self.sfactory.broadcast_line("test")
        return d

    def test_unknown_command(self):
        d = defer.Deferred()
        responses = ["Command not found: abracadabracadabr\\n", ]
        self.cfactory.receiver = self._get_expecting_line_receiver(
            responses, d)
        self.cfactory.message("abracadabracadabr")
        return d


def expected_join_responses(channel, callsign, ip, port):
    return [
        "socket channel '{0}' start creating: ip {1}:{2}\\n".format(
            channel, ip, port),
        "Chat: --- {0} joins the game.\\n".format(callsign),
        "socket channel '{0}', ip {1}:{2}, {3}, "
        "is complete created.\\n".format(
            channel, ip, port, callsign)]


def expected_leave_responses(channel, callsign, ip, port):
    return [
        "socketConnection with {0}:{1} on channel {2} lost.  "
        "Reason: \\n".format(
            ip, port, channel),
        "Chat: --- {0} has left the game.\\n".format(
            callsign)]

def expected_kick_responses(channel, callsign, ip, port):
    return [
        "socketConnection with {0}:{1} on channel {2} lost.  "
        "Reason: You have been kicked from the server.\\n".format(
            ip, port, channel),
        "Chat: --- {0} has left the game.\\n".format(
            callsign)]

def expected_load_responses(path):
    return [
        "Loading mission {0}...\\n".format(path),
        "Load bridges\\n",
        "Load static objects\\n",
        "##### House without collision (3do/Tree/Tree2.sim)\\n",
        "##### House without collision (3do/Buildings/Port/Floor/live.sim)\\n",
        "##### House without collision (3do/Buildings/Port/BaseSegment/"
            "live.sim)\\n",
        "Mission: {0} is Loaded\\n".format(path)]


class TestPilots(ConsoleBaseTestCase):

    def setUp(self):
        r = super(TestPilots, self).setUp()
        self.srvc = self.sservice.getServiceNamed('pilots')
        return r

    def tearDown(self):
        self.srvc = None
        return super(TestPilots, self).tearDown()

    def _get_pilots_count_checker(self, expected_count):
        def check(_):
            self.assertEqual(len(self.srvc.pilots), expected_count)
        return check

    def test_join(self):
        responses = expected_join_responses(
            1, "user1", "192.168.1.2", self.srvc.port)
        responses.extend(expected_join_responses(
            3, "user2", "192.168.1.3", self.srvc.port))

        d = defer.Deferred()
        d.addCallback(self._get_pilots_count_checker(2))
        self.cfactory.receiver = self._get_expecting_line_receiver(
            responses, d)

        self.srvc.join("user1", "192.168.1.2")
        self.srvc.join("user2", "192.168.1.3")
        return d

    def test_leave(self):
        responses = expected_join_responses(
            1, "user1", "192.168.1.2", self.srvc.port)
        responses.extend(expected_join_responses(
            3, "user2", "192.168.1.3", self.srvc.port))
        responses.extend(expected_leave_responses(
            1, "user1", "192.168.1.2", self.srvc.port))

        d = defer.Deferred()
        d.addCallback(self._get_pilots_count_checker(1))
        self.cfactory.receiver = self._get_expecting_line_receiver(
            responses, d)
        self.srvc.join("user1", "192.168.1.2")
        self.srvc.join("user2", "192.168.1.3")
        self.srvc.leave("user1")
        self.srvc.leave("fake_user")
        return d

    def test_kick(self):
        responses = expected_join_responses(
            1, "user1", "192.168.1.2", self.srvc.port)
        responses.extend(expected_join_responses(
            3, "user2", "192.168.1.3", self.srvc.port))
        responses.extend(expected_kick_responses(
            1, "user1", "192.168.1.2", self.srvc.port))

        d = defer.Deferred()
        d.addCallback(self._get_pilots_count_checker(1))
        self.cfactory.receiver = self._get_expecting_line_receiver(
            responses, d)
        self.srvc.join("user1", "192.168.1.2")
        self.srvc.join("user2", "192.168.1.3")
        self.cfactory.message("kick user1")
        return d


class TestMissions(ConsoleBaseTestCase):

    def setUp(self):
        r = super(TestMissions, self).setUp()
        self.srvc = self.sservice.getServiceNamed('missions')
        return r

    def tearDown(self):
        self.srvc = None
        return super(TestMissions, self).tearDown()

    def test_no_mission(self):
        d = defer.Deferred()
        responses = ["Mission NOT loaded\\n", ]
        self.cfactory.receiver = self._get_expecting_line_receiver(
            responses, d)
        self.cfactory.message("mission")
        return d

    def test_load_mission(self):
        responses = expected_load_responses("net/dogfight/test.mis")
        responses.append(responses[-1])

        d = defer.Deferred()
        self.cfactory.receiver = self._get_expecting_line_receiver(
            responses, d)
        self.cfactory.message("mission LOAD net/dogfight/test.mis")
        self.cfactory.message("mission")
        return d

    def test_begin_mission(self):
        responses = ["ERROR mission: Mission NOT loaded\\n"]
        responses.extend(expected_load_responses("net/dogfight/test.mis"))
        responses.append("Mission: net/dogfight/test.mis is Playing\\n")
        responses.append(responses[-1])

        d = defer.Deferred()
        self.cfactory.receiver = self._get_expecting_line_receiver(
            responses, d)
        self.cfactory.message("mission BEGIN")
        self.cfactory.message("mission LOAD net/dogfight/test.mis")
        self.cfactory.message("mission BEGIN")
        self.cfactory.message("mission")
        return d

    def test_end_mission(self):
        responses = ["ERROR mission: Mission NOT loaded\\n"]
        responses.extend(expected_load_responses("net/dogfight/test.mis"))
        responses.append("Mission: net/dogfight/test.mis is Playing\\n")
        responses.append("Mission: net/dogfight/test.mis is Loaded\\n")

        d = defer.Deferred()
        self.cfactory.receiver = self._get_expecting_line_receiver(
            responses, d)
        self.cfactory.message("mission END")
        self.cfactory.message("mission LOAD net/dogfight/test.mis")
        self.cfactory.message("mission BEGIN")
        self.cfactory.message("mission END")
        self.cfactory.message("mission END END END")
        self.cfactory.message("mission")
        return d

    def test_destroy_mission(self):
        responses = ["ERROR mission: Mission NOT loaded\\n"]
        responses.extend(expected_load_responses("net/dogfight/test.mis"))
        responses.extend(expected_load_responses("net/dogfight/test.mis"))
        responses.append("Mission: net/dogfight/test.mis is Playing\\n")
        responses.append("Mission NOT loaded\\n")

        d = defer.Deferred()
        self.cfactory.receiver = self._get_expecting_line_receiver(
            responses, d)
        self.cfactory.message("mission DESTROY")
        self.cfactory.message("mission LOAD net/dogfight/test.mis")
        self.cfactory.message("mission DESTROY")
        self.cfactory.message("mission LOAD net/dogfight/test.mis")
        self.cfactory.message("mission BEGIN")
        self.cfactory.message("mission DESTROY")
        self.cfactory.message("mission")
        return d



from setuptools import setup, find_packages

setup(
    name='il2ds-middleware',
    version='0.1.0',
    description='High-level access to IL-2 FB Dedicated Server.',
    long_description=open('README.md').read(),
    license='MIT',
    url='https://github.com/IL2HorusTeam/il2ds-middleware',
    bugtrack_url='https://github.com/IL2HorusTeam/il2ds-middleware/issues',
    author='Alexander Oblovatniy',
    author_email='oblovatniy@gmail.com',
    packages=find_packages(),
    install_requires=[i.strip() for i in open("requirements.pip").readlines()],
    classifiers = [
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries',
    ],
)
